"""Transparent and Interpretable Machine Learning Models in Python, with a keras inspired API

``lucidmode`` is a set of interpretable Machine Learning Models.

"""
__version__ = '0.4.3.1'

__all__ = ['functions', 'models', 'propagate', 'regularization']
